import datetime
import os
import re
import sys

from contextlib import contextmanager
from weakref import WeakValueDictionary


@contextmanager
def stderr_redirector(stream):
    old_stderr = sys.stderr
    sys.stderr = stream
    try:
        yield
    finally:
        sys.stderr = old_stderr


class SFile:
    @classmethod
    def mkdir(cls, fpath):
        if not os.path.exists(fpath):
            os.makedirs(fpath)

    @classmethod
    def to_file(cls, fpath, data):
        cls.mkdir(os.path.dirname(os.path.abspath(fpath)))
        with open(fpath, 'w') as fd:
            fd.write(data)

    @classmethod
    def read_all(cls, fpath):
        with open(fpath, 'r') as fd:
            return fd.read()

    @classmethod
    def readline(cls, fpath):
        with open(fpath, 'r') as fd:
            for line in fd:
                yield line
        # return ''


class SSingleton(type):
    _instances = {}  # WeakValueDictionary()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super(SSingleton, cls).__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class SStamp:
    DB_DATETIME = '%Y-%m-%d %H:%M:%S'
    DB_DATE = '%Y-%m-%d'
    DATE_SEPERATERS = '.-_/'

    @classmethod
    def now(cls, form=DB_DATETIME):
        return datetime.datetime.now().strftime(form)


class SStrExpand:
    MAX_LOOP_COUNT = 30

    class Error(Exception):
        pass

    # Reference: https://stackoverflow.com/a/30777398
    @classmethod
    def environ_vars(cls, string, default=None, skip_escaped=False):
        """Expand environment variables of form $var and ${var}.
           If parameter 'skip_escaped' is True, all escaped variable references
           (i.e. preceded by backslashes) are skipped.
           Unknown variables are set to 'default'. If 'default' is None,
           they are left unchanged.
        """
        def replace_var(m):
            defvalue = m.group(0) if default is None else default
            # print(m.group(2), m.group(1))
            return os.environ.get(m.group(2) or m.group(1), defvalue)

        reval = (r'(?<!\\)' if skip_escaped else '') + r'\$(\w+|\{([^}]*)\})'
        return re.sub(reval, replace_var, string)

    @classmethod
    def config_vars(cls, config, string):
        def replace_var(m):
            value = config.get_value(m.group(2) or m.group(1), m.group(0))
            if type(value) is list:
                # Just valid one-dimensional list
                return ','.join([str(x) for x in value])
            elif type(value) is dict:
                raise SStrExpand.Error('No Rules for dict.')
            return str(value)

        if config is None:
            return string
        reval = r'@([\w.]+|\{([^}]*)\})'
        return re.sub(reval, replace_var, string)

    @classmethod
    def convert(cls, string, config=None):
        lpcnt = 0
        o_string = string
        while True:
            p_string = string
            string = cls.config_vars(config, cls.environ_vars(p_string))
            # print(p_string, string)
            if p_string == string:
                return string
            lpcnt += 1
            if lpcnt > cls.MAX_LOOP_COUNT:
                emsg = 'Infinite Repeat Error: {}'.format(o_string)
                raise cls.Error(emsg)
        return string




import codecs
import os

from pysp.sbasic import SFile, SStamp


class STextFile(SFile):
    SPLIT_FILE_FORMAT = '{fname}-{idx:03d}.{ext}'
    EOL = '\r\n'

    def __init__(self, path):
        apath = path.split('.')
        self.fname = '.'.join(apath[:-1])
        self.ext = apath[-1]
        self.split_index = 0
        self.open_file(path)

    def __del__(self):
        if self.fd:
            self.fd.close()

    def generate_file_name(self, increment=0):
        self.split_index += increment
        return self.SPLIT_FILE_FORMAT.format(
                fname=self.fname, idx=self.split_index, ext=self.ext)

    def open_file(self, fname):
        self.mkdir(os.path.dirname(fname))
        self.curname = fname
        self.fd = codecs.open(fname, 'w', encoding='utf-8')

    def split_file(self, callback=None):
        self.fd.close()
        if self.split_index == 0:
            rename = self.generate_file_name(0)
            os.rename(self.curname, rename)
        self.open_file(self.generate_file_name(1))
        if callback:
            callback()

    def writeln(self, string):
        self.fd.write(string+self.EOL)


class SCSV(STextFile):
    MAX_FIELD_COUNT = 100000
    APPEND_SUBTITLE = '{title} / continue {idx:03d}'

    def __init__(self, path, title, columns):
        super(SCSV, self).__init__(path)
        self.title = title
        # self.schema = schema
        self.columns = columns
        self.title = title
        self.field_index = 0
        self.blank_line = ','*len(columns)
        self._need_split = False
        self.write_header()
        # for col in columns:
        #     if col not in schema:
        #         emsg = 'Not exists {col} in schema.'.format(col=col)
        #         raise PyJoyousError(DTAG, emsg)

    def write_header(self):
        if self.field_index == 0:
            _title = self.title
        else:
            _title = self.APPEND_SUBTITLE.format(
                                title=self.title, idx=self.split_index)
        self.writeln(_title + self.blank_line[:-1])
        self.writeln(SStamp.now() + self.blank_line[:-1])
        self.writeln(','+','.join(self.columns))

    def column_to_str(self, value):
        if type(value) is str and value.find('\n') >= 0:
            return '"'+value+'"'
        return '"=""'+str(value)+'"""'

    def write_field(self, field, to_str=False):
        if self._need_split:
            self._need_split = False
            self.split_file(self.write_header)

        self.field_index += 1
        if (self.field_index % self.MAX_FIELD_COUNT) == 0:
            self._need_split = True
        ftext = str(self.field_index)
        for i, v in enumerate(field):
            if to_str:
                ftext += ','+self.column_to_str(v)
            elif type(v) is str and v.find('\n') >= 0:
                ftext += ','+'"'+v+'"'
            else:
                ftext += ','+str(v)
        self.writeln(ftext)



# -*- coding: utf-8 -*-

import codecs
import copy
import collections
import os
import re
import yaml

from pysp.serror import SDebug, SError


YNode = collections.namedtuple('YNode', 'xpath fullpath value')


class SYAML(SDebug):
    TAG_INCLUDE = '!include'
    MARK_INCLUDE = '__include__'
    EOL = '\n'

    @classmethod
    def dump(cls, yo, pretty=True):
        if pretty:
            return yaml.dump(yo, default_flow_style=False, indent=4)
        return yaml.dump(yo)

    def merge(self, newy, defy):
        if isinstance(newy, dict) and isinstance(defy, dict):
            for k, v in defy.items():
                if k not in newy:
                    # self.dprint(k, v)
                    newy[k] = v
                else:
                    newy[k] = self.merge(newy[k], v)  # if newy[k] else v
        return newy

    def load(self, yml, node_value=None):
        yaml.add_constructor(SYAML.TAG_INCLUDE, self.include)
        if os.path.exists(yml):
            with codecs.open(yml, 'r', encoding='utf-8') as fd:
                try:
                    # pass a file descripter, if use '!include' method
                    return self.store_mark(yaml.load(fd), yml, node_value)
                    # return yaml.load(fd)
                except yaml.YAMLError as e:
                    emsg = 'Error YAML Loading: %s\n%s' % (yml, str(e))
                    raise SError(emsg)
        try:
            return yaml.load(yml)
        except yaml.YAMLError as e:
            emsg = 'Error YAML Loading: %s\n%s' % (yml, str(e))
            raise SError(emsg)

    def collect_node(self, ymlo, xpath=''):
        ns = []
        self.dprint("xpath: '{}'".format(xpath))
        for k, v in ymlo.items():
            xp = xpath
            self.dprint('- ', k)
            if k == self.MARK_INCLUDE:
                self.dprint('+ ', xp)
                n = YNode(xpath=xp,
                          fullpath=ymlo[k]['fullpath'],
                          value=ymlo[k]['value'])
                ns.append(n)
            elif type(ymlo[k]) == dict:
                xp += ('.' if xp else '') + k
                ns += self.collect_node(ymlo[k], xp)
        return ns

    def _check_folder(self, path):
        dirname = os.path.dirname(os.path.abspath(path))
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    def store(self, ymlo, pretty=True):
        '''Stored yml object to file.  In processing, yml object is changed.'''
        def store_file(ymlo, node):
            self._check_folder(node.fullpath)
            self.dprint('W: {}'.format(node.fullpath))
            with codecs.open(node.fullpath, 'w', encoding='utf-8') as fd:
                data = SYAML.dump(ymlo, pretty=pretty)
                for line in data.strip().splitlines():
                    if line.find(SYAML.TAG_INCLUDE) >= 0:
                        line = line.replace("'", "")
                    fd.write(line + SYAML.EOL)

        def get_node(ymlo, xpaths):
            ymlso = ymlo
            if xpaths and xpaths[0]:
                for k in xpaths:
                    ymlso = ymlso[k]
            return ymlso

        def store_node(ymlo, node):
            xpaths = node.xpath.split('.')
            ymlso = get_node(ymlo, xpaths)
            self.dprint('xpath={}, value={}'.format(node.xpath, node.value))
            self.dprint(SYAML.dump(ymlo, pretty=pretty))
            del ymlso[SYAML.MARK_INCLUDE]
            store_file(ymlso, node)

            ymlso = get_node(ymlo, xpaths[:-1])
            if node.value:
                ymlso[xpaths[-1]] = str(SYAML.TAG_INCLUDE + ' ' + node.value)
            self.dprint(SYAML.dump(ymlo, pretty=pretty))

        nodes = self.collect_node(ymlo)
        # self.DEBUG = True
        for n in nodes:
            self.dprint(f'{n.xpath}|{n.fullpath}|{n.value}')
            self.dprint('============')
            store_node(ymlo, n)

    def store_mark(self, yml, fullpath, node_value):
        if SYAML.MARK_INCLUDE in yml:
            emsg = 'Already use %s key in %s' % (SYAML.MARK_INCLUDE, fullpath)
            raise SError(emsg)
        yml[SYAML.MARK_INCLUDE] = {
            'fullpath': fullpath,
            'value': node_value
        }
        return yml

    def include(self, loader, node):
        # self.DEBUG = True
        fname = os.path.join(os.path.dirname(loader.name), node.value)
        self.dprint('Include YAML:', fname)
        return self.load(fname, node.value)


class SConfig(SYAML):
    _data = {}

    def __init__(self, yml_file=None):
        self._data = {}
        self._re_p = re.compile(r'([\w-]+)\[([-\d]+)\]')
        self._dirty = False
        if yml_file:
            self.loadup(yml_file)

    def dump(self):
        return super(SConfig, self).dump(self._data)

    def loadup(self, yml_file):
        if not os.path.exists(yml_file):
            raise SError('Not Exists: {}'.format(yml_file))
        if self._data:
            self.iprint('Overwrite')
        self._data = self.load(yml_file)

    def _fixup_folder(self, oyml, ymlpath):
        if ymlpath:
            nodes = self.collect_node(oyml)
            dirname = os.path.dirname(ymlpath)
            for n in nodes:
                if len(n.xpath.split('.')) > 1:
                    xpath = '{}.{}.fullpath'.format(n.xpath, self.MARK_INCLUDE)
                else:
                    xpath = '{}.fullpath'.format(self.MARK_INCLUDE)
                if n.value:
                    fpath = '{}/{}'.format(dirname, n.value)
                else:
                    fpath = ymlpath
                self.set_value(xpath, fpath, oyml=oyml)

    def collecting(self, yml_file):
        # if not os.path.exists(yml_file):
        #     raise SError('Not Exists: {}'.format(yml_file))
        if os.path.exists(yml_file):
            self._dirty = True
            newy = self.load(yml_file)
            self._data = self.merge(newy, self._data)
        self._fixup_folder(self._data, yml_file)

    def store(self, to_abs=None):
        data = copy.deepcopy(self._data)
        if to_abs:
            fullpath = os.path.abspath(to_abs)
            self.set_value(self.MARK_INCLUDE, {
                'fullpath': fullpath,
                'value': None
            }, oyml=data)
        self._fixup_folder(data, to_abs)
        if self._dirty:
            super(SConfig, self).store(data)

    def _parse_keylist(self, word):
        m = self._re_p.match(word)
        if m and m.lastindex == 2:
            return m.group(1), int(m.group(2))
        return None, 0

    def get_value(self, key, defvalue='', oyml=None):
        data = self._data if oyml is None else oyml
        for k in key.split('.'):
            if k in data:
                data = data[k]
                continue
            _k, _i = self._parse_keylist(k)
            if _k and _k in data and type(data[_k]) is list:
                try:
                    data = data[_k][_i]
                    continue
                except Exception:
                    pass
            return defvalue
        return data

    def set_value(self, key, value, oyml=None):
        self._dirty = True
        data = self._data if oyml is None else oyml
        karr = key.split('.')
        lenka = len(karr)
        for d, k in enumerate(karr):
            if k in data:
                if d == (lenka - 1):
                    data[k] = value
                else:
                    data = data[k]
                continue
            # keylist
            _k, _i = self._parse_keylist(k)
            if _k and _k in data and type(data[_k]) is list:
                if d == (lenka - 1):
                    try:
                        data[_k][_i] = value
                    except Exception:
                        # Append value to list,
                        # but it allows only the index is the next number
                        if len(data[_k]) == _i:
                            data[_k].append(value)
                            continue
                        raise SError('Index or Not Supported Format #1')
                else:
                    data = data[_k][_i]
                continue
            if _k:
                if _k not in data:
                    # Create a list value
                    if _i == 0:
                        data[_k] = []
                        data[_k].append(value)
                        continue
                raise SError('Index or Not Supported Format #2')
            # Others
            if d == (lenka - 1):
                data[k] = value
            else:
                data[k] = {}
                data = data[k]

    def delete(self, key, oyml=None):
        self._dirty = True
        data = self._data if oyml is None else oyml
        karr = key.split('.')
        lenka = len(karr)
        for d, k in enumerate(karr):
            if k in data:
                if d == (lenka - 1):
                    del data[k]
                else:
                    data = data[k]
                continue
            # keylist
            _k, _i = self._parse_keylist(k)
            if _k and _k in data and type(data[_k]) is list:
                if d == (lenka - 1):
                    try:
                        del data[_k][_i]
                    except IndexError:
                        raise SError('Index Out of Range')
                else:
                    data = data[_k][_i]
                continue
            # Others

    def dirty(self):
        if not self._dirty:
            self.dprint('Forcedly Set DIRTY')
        self._dirty = True



import os
import unittest

from pysp.sbasic import SStrExpand, SFile
from pysp.sconf import SConfig
from pysp.serror import SDebug


class StrExpandTest(SDebug, SFile, unittest.TestCase):

    def _set_environ(self, kv):
        if kv:
            idx = kv.index('=')
            if idx > 0:
                os.environ[kv[:idx]] = kv[idx+1:]

    def test_environ_vars(self):
        testcase = (
            (True,  'ABC', 'ABC',  ''),
            (True,  'ABC', '$KEY', 'KEY=ABC'),
            (True,  'ABC.Value', '$KEY.Value', 'KEY=ABC'),
            (True,  'MyABC.Value', 'My$KEY.Value', 'KEY=ABC'),
            (True,  'ABC', '${KEY}', 'KEY=ABC'),
            (True,  'ABC.Value', '${KEY}.Value', 'KEY=ABC'),
            (True,  'MyABC.Value', 'My${KEY}.Value', 'KEY=ABC'),
            (True,  'My$NoKEY.Value', 'My$NoKEY.Value', ''),
            (True,  'My${NoKEY}.Value', 'My${NoKEY}.Value', ''),
            (True,  'MyABC$Value', 'My$KEY$Value', 'KEY=ABC'),
            (True,  'MyABC$Value', 'My$KEY$Value', 'KEY=ABC'),
            (False, 'MyABC_Value', 'My$KEY_Value', 'KEY=ABC'),
            (True,  'MyABC_Value', 'My${KEY}_Value', 'KEY=ABC'),
            (True,  'MyABC123', 'My${KEY}$Value', 'Value=123'),
        )
        for sch in '!@#%^&*()-=+{}[]|,./;:\'"`~<>?':
            testcase += (
                (True,  'MyABC{sch}value'.format(sch=sch),
                        'My${KEY}{sch}value'.format(KEY='KEY', sch=sch), ''),
                (True,  'MyABC{sch}value'.format(sch=sch),
                        'My$KEY{sch}value'.format(sch=sch), ''),
            )

        for case in testcase:
            rv, expected, string, keyvalue = case
            if keyvalue:
                self._set_environ(keyvalue)
            # print('@@', SStrExpand.environ_vars(string))
            assert rv == (expected == SStrExpand.environ_vars(string))

    def test_config_vars(self):
        yml_string = '''
vehicle:
    sedan:
        fuel: [disel, gasoline]
        wheels: 4
    suv:
        fuel: [disel, gasoline]
        wheels: 4
'''
        vehicle_file = '/tmp/test/vehicle.yml'
        self.to_file(vehicle_file, yml_string)
        cfg = SConfig(vehicle_file)
        # self.DEBUG = True

        testcase = (
            (True,  cfg,
                    'Sedan has 4 tires.',
                    'Sedan has @vehicle.sedan.wheels tires.', ''),
            (True,  cfg,
                    'Engine is two types - disel,gasoline.',
                    'Engine is two types - @{vehicle.suv.fuel}.', ''),
            (True,  cfg,
                    'Engine is two types - @vehicle.suv.fuel.',
                    'Engine is two types - @vehicle.suv.fuel.', ''),
            (True,  None,
                    'Engine is two types - @{vehicle.suv.fuel}.',
                    'Engine is two types - @{vehicle.suv.fuel}.', ''),
        )

        for case in testcase:
            rv, config, expected, string, keyvalue = case
            if keyvalue:
                self._set_environ(keyvalue)

            assert rv == (expected == SStrExpand.config_vars(config, string))

    def test_convert(self):
        yml_string = '''
vehicle:
    sedan:
        fuel: [disel, gasoline]
        wheels: 4
    suv:
        fuel: [disel, gasoline]
        wheels: 4
link: $ENV_LINK
link2: "@link"
'''
        vehicle_file = '/tmp/test/vehicle.yml'
        self.to_file(vehicle_file, yml_string)
        cfg = SConfig(vehicle_file)
        # self.DEBUG = True

        testcase = (
            (True,  cfg,
                    'Sedan has 4 tires.',
                    'Sedan has @vehicle.sedan.wheels tires.', ''),
            (True,  cfg,
                    'Engine is two types - disel,gasoline.',
                    'Engine is two types - @{vehicle.suv.fuel}.', ''),
            (True,  cfg,
                    'Engine is two types - @vehicle.suv.fuel.',
                    'Engine is two types - @vehicle.suv.fuel.', ''),
            (True,  None,
                    'Engine is two types - @{vehicle.suv.fuel}.',
                    'Engine is two types - @{vehicle.suv.fuel}.', ''),
            (True,  cfg,
                    'All vehicle has not 4 tires.',
                    'All vehicle has not $VEHICLE_TIRES tires.',
                    'VEHICLE_TIRES=@{vehicle.sedan.wheels}'),
            (True,  cfg,
                    '$ENV_LINK',
                    '$ENV_LINK',
                    'ENV_LINK=@link'),
            (True,  None,
                    '@link',
                    '$ENV_LINK',
                    'ENV_LINK=@link'),
        )

        for case in testcase:
            rv, config, expected, string, keyvalue = case
            if keyvalue:
                self._set_environ(keyvalue)
            result = SStrExpand.convert(string, config=config)
            assert rv == (expected == result)

        self._set_environ('ENV_LINK=@link2')
        try:
            SStrExpand.convert('@link', config=cfg)
            assert False
        except SStrExpand.Error:
            pass




import shutil
import unittest

from pysp.sbasic import SFile
from pysp.serror import SDebug
from pysp.sconf import SConfig


class Expected:
    default_vehicle = '''
car:
    sedan:
        SM6:
            transmission: cvt
    suv:
        QM6:
            transmission: cvt
'''.strip()

    user_vehicle = '''
car:
    suv:
        QM6:
            color: cloud-perl
            transmission: CVT
'''.strip()

    overlay_user_vehicle = '''
car:
    sedan:
        SM6:
            transmission: cvt
    suv:
        QM6:
            color: cloud-perl
            transmission: CVT
'''.strip()


class FolderConfig:
    root = '''
car:
    sedan: !include folder/sedan
    suv: !include folder/suv

'''.strip()

    car_sedan = '''
SM6:
    transmission: cvt
'''.strip()

    car_suv = '''
QM6:
    transmission: cvt
'''.strip()


class ConfigList:
    target = '''
car:
    sedan:
      - name: SM6
        vendor: renault
      - name: Avante
        vendor: Hyundae
      - name: K7
        vendor: KIA
      - name: Others
        vendor:
          - Ford
          - BMW
          - Audi
'''.strip()


class ConfigTest(unittest.TestCase, SDebug, SFile):
    # DEBUG = True
    def_folder = '/tmp/yaml/default/'
    user_folder = '/tmp/yaml/user/'

    def config_create_default(self):
        cfg = SConfig()
        cfg.set_value('car.suv.QM6.transmission', 'cvt')
        cfg.set_value('car.sedan.SM6.transmission', 'cvt')
        cfgfile = self.def_folder+'vehicle'
        cfg.store(cfgfile)
        load_vehicle = self.read_all(cfgfile).strip()
        self.assertTrue(Expected.default_vehicle == load_vehicle)

    def config_create_user(self):
        cfg = SConfig()
        cfg.set_value('car.suv.QM6.transmission', 'CVT')
        cfg.set_value('car.suv.QM6.color', 'cloud-perl')
        cfgfile = self.user_folder+'vehicle'
        cfg.store(cfgfile)
        load_vehicle = self.read_all(cfgfile).strip()
        self.assertTrue(Expected.user_vehicle == load_vehicle)

    def config_merge(self):
        cfg = SConfig(self.def_folder+'vehicle')
        cfg.collecting(self.user_folder+'vehicle')
        cfg.store()
        cfgfile = self.user_folder+'vehicle'
        load_vehicle = self.read_all(cfgfile).strip()
        self.assertTrue(Expected.overlay_user_vehicle == load_vehicle)

    def test_config(self):
        self.config_create_default()
        self.config_create_user()
        self.config_merge()

    def test_config_folder(self):
        shutil.rmtree(self.def_folder)
        shutil.rmtree(self.user_folder)
        self.to_file(self.def_folder+'vehicle', FolderConfig.root)
        self.to_file(self.def_folder+'/folder/sedan', FolderConfig.car_sedan)
        self.to_file(self.def_folder+'/folder/suv', FolderConfig.car_suv)
        cfg = SConfig(self.def_folder+'vehicle')
        # self.DEBUG = True
        self.dprint('\n'+cfg.dump())
        cfg.collecting(self.user_folder+'vehicle')
        self.dprint('\n'+cfg.dump())
        cfg.store()
        cfg2 = SConfig(self.user_folder+'vehicle')
        self.dprint('\n'+cfg.dump())
        self.dprint('\n'+cfg2.dump())
        self.assertTrue(cfg.dump() == cfg2.dump())
        for fname in ['vehicle', 'folder/sedan', 'folder/suv']:
            default_load = self.read_all(self.def_folder+fname).strip()
            user_load = self.read_all(self.user_folder+fname).strip()
            self.assertTrue(default_load == user_load)

    def test_access_keylist(self):
        cfgfile = '/tmp/yaml2/config_list.yml'
        self.to_file(cfgfile, ConfigList.target)
        cfg = SConfig(cfgfile)
        self.assertTrue(cfg.get_value('car.sedan[2].name') == 'K7')
        self.assertTrue(cfg.get_value('car.sedan[1].vendor') == 'Hyundae')
        self.assertTrue(cfg.get_value('car.sedan[0].name') == 'SM6')
        self.assertTrue(cfg.get_value('car.sedan[-2].name') == 'K7')
        self.assertTrue(cfg.get_value('car.sedan[100].name', 'NA') == 'NA')
        self.assertTrue(cfg.get_value('car.sedan[-1].vendor[1]') == 'BMW')
        cfg.set_value('car.sedan[-1].name', 'Overseas')
        self.assertTrue(cfg.get_value('car.sedan[-1].name') == 'Overseas')
        cfg.set_value('car.sedan[-1].vendor[-1]', 'Benz')
        self.assertTrue(cfg.get_value('car.sedan[-1].vendor[-1]') == 'Benz')
        # Append value to list, but it allows only the index is the next number
        cfg.set_value('car.sedan[-1].vendor[3]', 'Audi')
        self.assertTrue(cfg.get_value('car.sedan[-1].vendor[-1]') == 'Audi')
        cfg.set_value('car.suv[0]', 'QM6')
        cfg.set_value('car.suv[1]', 'Tusan')
        cfg.set_value('car.suv[2]', 'Tivoli')
        self.assertTrue(cfg.get_value('car.suv[-1]') == 'Tivoli')

        cfg.delete('car.suv[1]')
        self.assertTrue(cfg.get_value('car.suv[1]') == 'Tivoli')
        cfg.delete('car.sedan[-1].vendor[-1]')
        self.assertTrue(cfg.get_value('car.sedan[-1].vendor[-1]') == 'Benz')
        cfg.set_value('car.suv', None)
        self.assertTrue(cfg.get_value('car.suv') is None)
        cfg.delete('car.suv')
        self.assertTrue(cfg.get_value('car.suv', 'Nothing') == 'Nothing')
        cfg.delete('car.sedan[-1].vendor')
        result = cfg.get_value('car.sedan[-1].vendor', 'Nothing')
        self.assertTrue(result == 'Nothing')
        cfg.store()




import datetime
import os
import shutil
import sqlalchemy as sa
import unittest

from pysp.sbasic import SFile
from pysp.serror import SDebug
from pysp.sconf import SConfig
from pysp.ssql import SSimpleDB


class SsqlTest(unittest.TestCase, SDebug, SFile):
    # DEBUG = True
    test_folder = '/tmp/sql/'
    config_file = test_folder+'db.cfg'
    db_file = test_folder+'test.sqlite3'

    def test_db(self):
        if os.path.exists(self.test_folder):
            shutil.rmtree(self.test_folder)
        try:
            self.check_create()
            self.check_alter()
            self.check_upsert()
            self.check_query()
            self.check_count()
            self.check_drop_columns()
            self.check_rename_columns()
        except Exception:
            self.assertTrue(False)

    def check_create(self):
        DBCONFIG = '''
tables:
    - name: inspection
      columns:
        - [SNumber, String10, NotNull, CollateNocase, Unique, PrimaryKey]
        - [Integer, Integer, NotNull]
        - [Float, Float]
        - [Bool, Boolean]
        - [DateTime, DateTime]
'''
        self.to_file(self.config_file, DBCONFIG)
        dbconfig = SConfig(self.config_file)
        db = SSimpleDB(self.db_file, dbconfig)
        del db

    def check_alter(self):
        DBCONFIG = '''
tables:
    - name: inspection
      columns:
        - [SNumber, String10, NotNull, CollateNocase, Unique, PrimaryKey]
        - [Integer, Integer, NotNull]
        - [Float, Float]
        - [Bool, Boolean]
        - [DateTime, DateTime]
        - [Date, Date]
        - [Alter1, String15, Unique]
        - [Alter2, String5]
'''
        self.to_file(self.config_file, DBCONFIG)
        db = SSimpleDB(self.db_file, SConfig(self.config_file))
        del db

    def check_upsert(self):
        db = SSimpleDB(self.db_file, SConfig(self.config_file))
        data = {
            'SNumber':  '0000000000',
            'Integer':  1234,
            'Date':     datetime.datetime.now().date(),
        }
        items = {
            'data': data
        }
        db.upsert('inspection', **items)
        data = {
            'SNumber':  '0000000000',
            'Alter2':   'Five5',
            'DateTime': sa.sql.func.now(),
        }
        items = {
            'data': data
        }
        db.upsert('inspection', **items)
        for i in range(10):
            data = {
                'SNumber':  '%010d' % i,
                'Integer':  9 - i,
                'Alter1':   '%05d' % i,
            }
            items = {
                'data': data
            }
            db.upsert('inspection', **items)
        del db

    def check_query(self):
        db = SSimpleDB(self.db_file, SConfig(self.config_file))
        columns = ['SNumber', 'Integer', 'Float', 'Alter1', 'Alter2']
        options = {
            'wheres': {
                'Alter1': ['00000', '00005'],
                'Alter2': 'Five5',
            },
            'operate': SSimpleDB.OP_AND,
            'page': 0,
            'size': 4,
        }
        data = db.query('inspection', *columns, **options)
        self.assertTrue(len(data) == 1)
        data = data[0]
        self.assertTrue(data[0] == '0000000000')
        self.assertTrue(data[1] == 9)
        self.assertTrue(data[2] is None)
        self.assertTrue(data[3] == '00000')
        self.assertTrue(data[4] == 'Five5')

        columns = []
        options = {
            'wheres': {
                'Alter1': ['00000', '00005'],
                'Alter2': 'Five5',
            },
            'operate': SSimpleDB.OP_OR,
            'page': 0,
            'size': 4,
        }
        data = db.query('inspection', *columns, **options)
        self.assertTrue(len(data) == 2)
        for item in data:
            self.assertTrue(item[0] in ['0000000000', '0000000005'])

    def check_count(self):
        db = SSimpleDB(self.db_file, SConfig(self.config_file))
        self.assertTrue(db.count('inspection') == 10)
        columns = []
        options = {
            'wheres': {
                'Alter1': ['00000', '005'],
                'Alter2': 'Five5',
            },
            'operate': SSimpleDB.OP_OR,
            'page': 0,
            'size': 4,
        }
        self.assertTrue(db.count('inspection', *columns, **options) == 2)
        columns = ['SNumber', 'Integer', 'Float', 'Alter1', 'Alter2']
        options = {
            'wheres': {
                'Alter1': ['00000', '00005'],
                'Alter2': 'Five5',
            },
            'operate': SSimpleDB.OP_AND,
            'page': 0,
            'size': 4,
        }
        self.assertTrue(db.count('inspection', *columns, **options) == 1)

    def check_drop_columns(self):
        DBCONFIG = '''
tables:
    - name: inspection
      columns:
        - [SNumber, String10, NotNull, CollateNocase, Unique, PrimaryKey]
        - [Integer, Integer, NotNull]
        - [Float, Float]
        - [DateTime, DateTime]
        - [Date, Date]
        - [Alter2, String5]
      migrate:
        operation: drop
'''
        self.to_file(self.config_file, DBCONFIG)
        db = SSimpleDB(self.db_file, SConfig(self.config_file))
        columns = ['SNumber', 'Integer', 'Alter2']
        options = {
            'wheres': {
                'Alter2': 'Five5',
            },
            'operate': SSimpleDB.OP_AND,
            'page': 0,
            'size': 4,
        }
        self.assertTrue(db.count('inspection', *columns, **options) == 1)
        del db

    def check_rename_columns(self):
        DBCONFIG = '''
tables:
    - name: inspection
      columns:
        - [SNumber, String10, NotNull, CollateNocase, Unique, PrimaryKey]
        - [Pass, String5]
        - [AString, String20]
        - [Integer, Integer, NotNull]
        - [DateTime, DateTime]
        - [Date, Date]
      migrate:
        operation: rename
        columns:
            Pass: Alter2
'''
        self.to_file(self.config_file, DBCONFIG)
        db = SSimpleDB(self.db_file, SConfig(self.config_file))
        columns = ['SNumber', 'Integer', 'Pass']
        options = {
            'wheres': {
                'Pass': 'Five5',
            },
            'operate': SSimpleDB.OP_AND,
            'page': 0,
            'size': 4,
        }
        self.assertTrue(db.count('inspection', *columns, **options) == 1)
        db.vacuum()
        del db



# -*- coding: utf-8 -*-

import codecs
import hexdump
import unittest
import os

from pysp.scsv import SCSV


NORMAL_CONTENT = '''Normal,,\r
2018-10-30 22:37:02,,\r
,string,integer,float\r
1,text,1,1.1\r
2,"t
e
x
t
",2,2.2\r
'''

STRING_CONTENT = '''To String,,\r
2018-10-30 23:06:45,,\r
,string,integer,float\r
1,"=""text""","=""1""","=""1.1"""\r
2,"t
e
x
t
","=""2""","=""2.2"""\r
'''


class CSVTest(unittest.TestCase):
    TMP_FOLDER = '/tmp/csv/'
    TMP_CSV_NORMAL_FILE = TMP_FOLDER+'csv_normal.csv'
    TMP_CSV_STRING_FILE = TMP_FOLDER+'csv_string.csv'
    TMP_CSV_SPLIT_FILE = TMP_FOLDER+'csv_split.csv'

    data_columns = ['string', 'integer', 'float']
    data_fields = [
            ['text', 1, 1.1],
            ['t\ne\nx\nt\n', 2, 2.2],
        ]

    # def tearDown(self):
    #     self.delete_file(self.TMP_CSV_NORMAL_FILE)
    #     self.delete_file(self.TMP_CSV_STRING_FILE)
    #
    # def delete_file(self, del_file):
    #     if os.path.exists(del_file):
    #         os.remove(del_file)

    def test_normal_csv_file(self):
        csv = SCSV(self.TMP_CSV_NORMAL_FILE, 'Normal', self.data_columns)
        for field in self.data_fields:
            csv.write_field(field)
        del csv
        with codecs.open(self.TMP_CSV_NORMAL_FILE, 'r', encoding='utf-8') as f:
            content = f.read()
            hexdump.hexdump(NORMAL_CONTENT.encode('utf-8'))
            hexdump.hexdump(content.encode('utf-8'))
            self.assertTrue(content[0x20:] == NORMAL_CONTENT[0x20:])

    def test_string_csv_file(self):
        csv = SCSV(self.TMP_CSV_STRING_FILE, 'To String', self.data_columns)
        for field in self.data_fields:
            csv.write_field(field, to_str=True)
        del csv
        with codecs.open(self.TMP_CSV_STRING_FILE, 'r', encoding='utf-8') as f:
            content = f.read()
            hexdump.hexdump(STRING_CONTENT.encode('utf-8'))
            hexdump.hexdump(content.encode('utf-8'))
            self.assertTrue(content[0x20:] == STRING_CONTENT[0x20:])

    def test_split_csv_file(self):
        field = ['split', 3, 3.333]
        SCSV.MAX_FIELD_COUNT = 10
        csv = SCSV(self.TMP_CSV_SPLIT_FILE, 'SPLIT CSV', self.data_columns)
        for i in range(25):
            csv.write_field(field)
        del csv

        apath = self.TMP_CSV_SPLIT_FILE.split('.')
        fname = '.'.join(apath[:-1])
        ext = apath[-1]
        for i in range(3):
            filename = SCSV.SPLIT_FILE_FORMAT.format(
                        fname=fname, idx=i, ext=ext)
            print(filename)
            self.assertTrue(os.path.exists(filename))



#!/usr/bin/env python3

import copy
import sqlalchemy as sa

from sqlalchemy import or_, and_  # , event
# from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker


# @event.listens_for(Engine, "connect")
# def set_sqlite_pragma(dbapi_connection, connection_record):
#     module = dbapi_connection.__class__.__module__
#     if module == 'sqlite3':
#         print('------------------------------------------->>>')
#         cursor = dbapi_connection.cursor()
#         cursor.execute("PRAGMA journal_mode=WAL")
#         cursor.execute("PRAGMA foreign_keys=ON")
#         cursor.close()


class SSQL(object):
    TEMP_PREFIX = 'tmp_'

    class Error(Exception):
        pass

    def __init__(self, config):
        self.meta = sa.MetaData()
        self.engine = None
        self.config = config

    def init_tables(self):
        meta = sa.MetaData(bind=self.engine)
        meta.reflect()
        for idx, dictable in enumerate(self.config.get_value('tables')):
            self._init_table(meta, dictable)

    def _init_table(self, meta, dictable):
        if dictable['name'] not in meta.tables.keys():
            self._create_table(meta, dictable)
            meta.create_all(self.engine)
        else:
            self._init_columns(meta, dictable)
            if 'migrate' in dictable and 'operation' in dictable['migrate']:
                operation = dictable['migrate']['operation']
                if operation == 'drop':
                    self.drop_columns(meta, dictable)
                elif operation == 'rename':
                    self.rename_columns(meta, dictable)

        if 'migrate' in dictable:
            self.config.delete('migrate', dictable)

    def _create_table(self, meta, dictable):
        tablename = dictable['name']
        columns = dictable['columns']
        args = [tablename, meta]
        for colparams in columns:
            args.append(self.build_column(colparams))
        sa.Table(*args)

    def _init_columns(self, meta, dictable):
        tablename = dictable['name']
        columns = dictable['columns']
        for colparam in columns:
            colname = colparam[0]
            if colname in meta.tables[tablename].columns:
                column = meta.tables[tablename].columns[colname]
                coltype = column.type.__class__.__name__
                if not self.is_valid_type(colparam[1], coltype):
                    emsg = 'Column {cn}: Not Matched Type - {a}, {b}'.format(
                        cn=colname, a=colparam[1], b=coltype)
                    raise SSQL.Error(emsg)
                if coltype in ['VARCHAR', 'CHAR']:
                    coltype = column.type
                    # print('@@@', colname, coltype.collation, coltype.length)
                    length = int(colparam[1][len('String'):])
                    if length != coltype.length:
                        emsg = 'Column {cn}: Not Matched String Length::' \
                            'Expected %d, but %d' % (length, coltype.length)
                        raise SSQL.Error(emsg.format(cn=colname))
                if 'PrimaryKey' in colparam and column.primary_key is False:
                    emsg = 'Column {cn}: Not Set Primary Key Property'
                    raise SSQL.Error(emsg.format(cn=colname))
                if 'Unique' in colparam and column.unique is False:
                    emsg = 'Column {cn}: Not Set Unique Property'
                    raise SSQL.Error(emsg.format(cn=colname))
                if 'NotNull' in colparam and column.nullable is True:
                    emsg = 'Column {cn}: Not Null'
                    raise SSQL.Error(emsg.format(cn=colname))
            else:
                colobj = self.build_column(colparam)
                self.add_column(self.engine, tablename, colobj)

    def is_valid_type(cls, config_type, sql_type):
        dbtypes = {
            'INTEGER':  'Integer',
            'FLOAT':    'Float',
            'BOOLEAN':  'Boolean',
            'DATETIME': 'DateTime',
            'DATE':     'Date',
        }
        if sql_type in ['VARCHAR', 'CHAR']:
            return (config_type.find('String') == 0)
        return dbtypes.get(sql_type) == config_type

    def build_column(self, params):
        dbtypes = {
            'Boolean':  sa.Boolean,
            'DateTime': sa.DateTime,
            'Float':    sa.Float,
            'Integer':  sa.Integer,
            'String':   sa.String,
            'Date':     sa.Date,
        }

        def build_type(coltype, params):
            if coltype.find('String') == 0:
                args = []
                kwargs = {}
                if len('String') == len(coltype):
                    raise SSQL.Error('Need Length of String')
                args.append(int(coltype[len('String'):]))
                if 'CollateNocase' in params:
                    kwargs['collation'] = 'NOCASE'
                return sa.String(*args, **kwargs)
            else:
                try:
                    return dbtypes.get(coltype)()
                except Exception:
                    raise SSQL.Error('Unknown Column Type: {}'.format(coltype))

        params = copy.deepcopy(params)
        args = []
        kwargs = {}
        args.append(params.pop(0))
        coltype = params.pop(0)
        args.append(build_type(coltype, params))
        kwargs['nullable'] = False if 'NotNull' in params else True
        kwargs['primary_key'] = True if 'PrimaryKey' in params else False
        kwargs['unique'] = True if 'Unique' in params else False
        # print(str(params), str(args), str(kwargs))
        return sa.Column(*args, **kwargs)

    def get_table(self, tablename):
        return sa.Table(tablename, self.meta, autoload=True,
                        autoload_with=self.engine)

    def add_column(self, engine, tablename, column):
        colname = column.compile(dialect=engine.dialect)
        coltype = column.type.compile(engine.dialect)
        sql = 'ALTER TABLE %s ADD COLUMN %s %s' % (tablename, colname, coltype)
        # print(sql)
        engine.execute(sql)

    def _drop_columns(self, meta, dictable, colnames):
        # BEGIN TRANSACTION;
        # ## XXX: This function worked, but lost each the columns property !!
        # ## CREATE TABLE t1_backup AS SELECT a, b FROM t1;
        # CREATE TABLE t1_backup ( columns with property )
        # INSERT INTO t1_backup SELECT a, b FROM t1;
        # DROP TABLE t1;
        # ALTER TABLE t1_backup RENAME TO t1;
        # COMMIT;
        tablename = dictable['name']
        dictable['name'] = self.TEMP_PREFIX + tablename
        sqls = ['INSERT INTO tmp_{tn} SELECT {cns} FROM {tn}',
                'DROP TABLE {tn}',
                'ALTER TABLE tmp_{tn} RENAME TO {tn}']
        form = {
            'tn':   tablename,
            'ttn':  dictable['name'],
            'cns':  ','.join(colnames)
        }
        # Check to exist tmp_tablename
        try:
            self.get_table(form['ttn'])
            emsg = 'Already Exists Table "{ttn}"'
            raise SSQL.Error(emsg.format(**form))
        except sa.exc.NoSuchTableError:
            self._create_table(meta, dictable)
            meta.create_all(self.engine)

        self.session.begin_nested()
        try:
            for sql in sqls:
                self.session.execute(sql.format(**form))
            self.session.commit()
        except Exception:
            self.session.rollback()
        # Restore the name of table
        dictable['name'] = tablename

    def drop_columns(self, meta, dictable):
        '''It handles columns of a table to the process of dropping and adding,
        concurrently.'''
        remains = []
        for colparam in dictable['columns']:
            remains.append(colparam[0])
        self._drop_columns(meta, dictable, remains)

    def rename_columns(self, meta, dictable):
        '''It handles columns of a table to the process of dropping, renaming,
        and adding, concurrently.'''
        remains = []
        renpool = dictable['migrate']['columns']
        for colparam in dictable['columns']:
            if colparam[0] in renpool:
                remains.append(renpool[colparam[0]])
            else:
                remains.append(colparam[0])
        self._drop_columns(meta, dictable, remains)


class SSimpleDB(SSQL):
    SQL_ECHO = False
    OP_AND = 'and'
    OP_OR = 'or'
    SQL_PAGE = 'page'
    SQL_SIZE = 'size'
    SQL_V_PAGE = 0
    SQL_V_SIZE = 5

    def __init__(self, dbpath, config):
        super(SSimpleDB, self).__init__(config)
        self.engine = sa.create_engine(
            'sqlite:///{db}'.format(db=dbpath), echo=self.SQL_ECHO)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()
        if self.session.bind.dialect.name == 'sqlite':
            self.session.execute("PRAGMA journal_mode=WAL")
            self.session.execute("PRAGMA foreign_keys=ON")
            self.session.commit()
        self.init_tables()

    def __del__(self):
        if hasattr(self, 'session'):
            self.session.close()
        if self.config and hasattr(self.config, 'store'):
            self.config.store()

    def to_sql(self, o):
        return str(o.compile(compile_kwargs={"literal_binds": True}))

    def _get_columns(self, table, colnames):
        columns = []
        for col in colnames:
            if col in table.c:
                columns.append(table.c[col])
        return columns if columns else table.c

    def get_colnames(self, tablename):
        table = self.get_table(tablename)
        return [x.name for x in table.c]

    def _append_wheres(self, qo, table, **kwargs):
        wheres = kwargs.pop('wheres', {})
        operate = kwargs.pop('operate', self.OP_AND)
        _op = {
            self.OP_AND:    and_,
            self.OP_OR:     or_,
        }
        if wheres:
            filters = []
            for k, v in wheres.items():
                # column = sa.sql.column(k)
                column = table.c[k]
                is_str = column.type.__class__.__name__ in ['VARCHAR', 'CHAR']
                if type(v) is list:
                    _o = None
                    for _v in v:
                        if _o is None:
                            if is_str:
                                _o = or_(column.like('%{}%'.format(_v)))
                            else:
                                _o = or_(column == _v)
                        else:
                            if is_str:
                                _o = or_(column.like('%{}%'.format(_v)), _o)
                            else:
                                _o = or_(column == _v, _o)
                    if _o is not None:
                        filters.append(_o)
                else:
                    if is_str:
                        colike = column.like('%{}%'.format(v))
                        filters.append(_op[operate](colike))
                    else:
                        filters.append(_op[operate](column == v))
            qo = qo.where(_op[operate](*filters))
        return qo

    def _build_query(self, table, *args, **kwargs):
        columns = self._get_columns(table, args)

        qo = sa.sql.select(columns)
        qo = self._append_wheres(qo, table, **kwargs)
        return qo

    def upsert(self, tablename, **kwargs):
        tbl = self.get_table(tablename)
        data = kwargs.get('data')
        only_insert = kwargs.get('only_insert', False)
        rv = False
        pk = next(iter(data))
        # pk_column = sa.sql.column(pk)
        pk_column = tbl.c[pk]

        qc = sa.sql.select([tbl.c[pk]]).where(pk_column == data[pk])
        qi = sa.insert(tbl).values(**data)
        qu = sa.update(tbl).values(**data).where(pk_column == data[pk])
        self.session.begin_nested()
        try:
            item = self.session.query(qc).first()
            if item is None:
                self.session.execute(qi)
                rv = True
            else:
                if only_insert:
                    rv = False
                else:
                    self.session.execute(qu)
                    rv = True
            self.session.commit()
        except Exception:
            self.session.rollback()
        return rv

    def upsert_array(self, tablename, **kwargs):
        '''
        param data - If the content of the array had mixed in the 'only_insert'
                     mode, it updates only non-existent data.
        '''
        tbl = self.get_table(tablename)
        arr_data = kwargs.get('data')
        only_insert = kwargs.get('only_insert', False)

        executed_count = 0
        self.session.begin_nested()
        for data in arr_data:
            pk = next(iter(data))
            pk_column = tbl.c[pk]
            qc = sa.sql.select([tbl.c[pk]]).where(pk_column == data[pk])
            qi = sa.insert(tbl).values(**data)
            qu = sa.update(tbl).values(**data).where(pk_column == data[pk])
            try:
                item = self.session.query(qc).first()
                if item is None:
                    self.session.execute(qi)
                    executed_count += 1
                else:
                    if only_insert:
                        continue
                    else:
                        self.session.execute(qu)
                        executed_count += 1
            except Exception:
                self.session.rollback()
                return False
        self.session.commit()
        return True if executed_count > 0 else False

    def query(self, tablename, *args, **kwargs):
        size = kwargs.pop(self.SQL_SIZE, self.SQL_V_SIZE)
        page = kwargs.pop(self.SQL_PAGE, self.SQL_V_PAGE)

        tbl = self.get_table(tablename)
        qo = self._build_query(tbl, *args, **kwargs)
        qo = qo.limit(size)
        qo = qo.offset(page*size)

        try:
            return self.session.query(qo).all()
        except Exception:
            raise SSimpleDB.Error('Query: {}'.format(self.to_sql(qo)))

    def count(self, tablename, *args, **kwargs):
        tbl = self.get_table(tablename)
        qo = self._build_query(tbl, *args, **kwargs)
        try:
            return self.session.query(qo).count()
        except Exception:
            raise SSimpleDB.Error('Count: {}'.format(self.to_sql(qo)))

    def vacuum(self):
        if self.session.bind.dialect.name == 'sqlite':
            self.session.execute('VACUUM')
            return
        raise SSimpleDB.Error('Not Implemented to VACCUM')



# -*- coding: utf-8 -*-

import os
import shutil
import unittest

from pysp.sbasic import SFile
from pysp.serror import SDebug
from pysp.sconf import SYAML


main_yml = '''
main:
    cmd: main-command
    sublevel: !include sub_l1.yml
    tag: main
'''.strip()

sub_l1_yml = '''
cmd: sub-command
sublevel: !include sub_l2.yml
tag: sub level 1
'''.strip()

sub_l2_yml = '''
cmd: sub-command
sublevel: null
tag: sub level 2
'''.strip()

expected_mark_yml = '''
__include__:
    fullpath: /tmp/yaml/main.yml
    value: null
main:
    cmd: main-command
    sublevel:
        __include__:
            fullpath: /tmp/yaml/sub_l1.yml
            value: sub_l1.yml
        cmd: sub-command
        sublevel:
            __include__:
                fullpath: /tmp/yaml/sub_l2.yml
                value: sub_l2.yml
            cmd: sub-command
            sublevel: null
            tag: sub level 2
        tag: sub level 1
    tag: main
'''.strip()

merge_yml = '''
main:
    sublevel:
        sublevel:
            sublevel: Merge Point 1
merge: Merge Point 2
'''.strip()

expected_merge_yml = '''
__include__:
    fullpath: /tmp/yaml/main.yml
    value: null
main:
    cmd: main-command
    sublevel:
        __include__:
            fullpath: /tmp/yaml/sub_l1.yml
            value: sub_l1.yml
        cmd: sub-command
        sublevel:
            __include__:
                fullpath: /tmp/yaml/sub_l2.yml
                value: sub_l2.yml
            cmd: sub-command
            sublevel: Merge Point 1
            tag: sub level 2
        tag: sub level 1
    tag: main
merge: Merge Point 2
'''.strip()

yml_files = [main_yml, sub_l1_yml, sub_l2_yml]


def get_var_name(var, dir=locals()):
    return [key for key, val in dir.items() if id(val) == id(var)][0]


class YamlTest(unittest.TestCase, SDebug, SFile):
    # DEBUG = True
    folder = '/tmp/yaml/'

    def convert_filename(self, var):
        fname = get_var_name(var).split('_')
        return self.folder+'_'.join(fname[:-1])+'.'+fname[-1]

    def load_test(self, yaml):
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)

        for item in yml_files:
            fpath = self.convert_filename(item)
            self.to_file(fpath, item)
        yfile = self.convert_filename(yml_files[0])
        self.ymlo = yaml.load(yfile)
        loaded_yml = yaml.dump(self.ymlo, pretty=True).strip()
        self.dprint(loaded_yml)
        self.assertTrue(loaded_yml == expected_mark_yml)

    def store_test(self, yaml):
        shutil.rmtree(self.folder)
        os.makedirs(self.folder)
        yaml.store(self.ymlo)

        for item in yml_files:
            fpath = self.convert_filename(item)
            data = self.read_all(fpath).strip()
            self.dprint('Load: {}'.format(fpath))
            self.assertTrue(data == item)

    def test_yaml(self):
        yaml = SYAML()
        self.load_test(yaml)
        self.store_test(yaml)

    def test_yaml_merge(self):
        yaml = SYAML()
        self.load_test(yaml)
        dy = self.ymlo
        uy = yaml.load(merge_yml)
        yo = yaml.merge(uy, dy)
        self.dprint(yaml.dump(yo, pretty=True))
        loaded_yml = yaml.dump(yo, pretty=True).strip()
        self.assertTrue(loaded_yml == expected_merge_yml)



# from __future__ import print_function
import logging
import os
# import sys

from inspect import getframeinfo, stack


_log = logging.getLogger('pysp')
_log.handlers = []
if 'DEBUG_PYTHON' in os.environ:
    _log_level = logging.DEBUG
    _log.setLevel(_log_level)
    h = logging.StreamHandler()
    h.setLevel(_log_level)
    _log.addHandler(h)
    # _log.propagate = 0

# _LOG_FORMAT = '%(asctime)s: %(message)s'
# _lhandler = logging.StreamHandler()
# _lhandler.setFormatter(logging.Formatter(_LOG_FORMAT))
# _log.addHandler(_lhandler)
# _log.propagate = 0


class Debug:
    DEBUG = False
    TAG_DEBUG = '[D] '
    TAG_ERROR = '[E] '
    TAG_INFO = '[I] '

    @classmethod
    def _print(cls, tag, *args):
        log_print = {
            cls.TAG_DEBUG: _log.debug,
            cls.TAG_ERROR: _log.error,
            cls.TAG_INFO: _log.info,
        }
        caller = getframeinfo(stack()[2][0])
        filename = os.path.basename(caller.filename)
        lmsg = f'{caller.lineno:4d}:{filename:<20}{tag} {" ".join(args)}'
        log_print.get(tag)(lmsg)


class SDebug(Debug):

    def dprint(self, *args, **kwargs):
        if self.DEBUG:
            self._print(self.TAG_DEBUG, *args)
            # _log.debug(f"{self.TAG_DEBUG} {' '.join(args)}")
            # print(self.TAG_DEBUG, *args, file=sys.stderr, **kwargs)

    def eprint(self, *args, **kwargs):
        self._print(self.TAG_ERROR, *args)
        # _log.error(f"{self.TAG_ERROR} {' '.join(args)}")
        # print(self.TAG_ERROR, *args, file=sys.stderr, **kwargs)

    def iprint(self, *args, **kwargs):
        self._print(self.TAG_INFO, *args)
        # _log.info(f"{self.TAG_INFO} {' '.join(args)}")
        # print(self.TAG_INFO, *args, file=sys.stderr, **kwargs)


class SCDebug(Debug):
    @classmethod
    def dprint(cls, *args, **kwargs):
        if cls.DEBUG:
            cls._print(cls.TAG_DEBUG, *args)
            # _log.debug(f"{cls.TAG_DEBUG} {' '.join(args)}")
            # print(cls.TAG_DEBUG, *args, file=sys.stderr, **kwargs)

    @classmethod
    def eprint(cls, *args, **kwargs):
        cls._print(cls.TAG_ERROR, *args)
        # _log.error(f"{cls.TAG_ERROR} {' '.join(args)}")
        # print(cls.TAG_ERROR, *args, file=sys.stderr, **kwargs)

    @classmethod
    def iprint(cls, *args, **kwargs):
        cls._print(cls.TAG_INFO, *args)
        # _log.info(f"{cls.TAG_INFO} {' '.join(args)}")
        # print(cls.TAG_INFO, *args, file=sys.stderr, **kwargs)


class SError(Exception):
    pass


if __name__ == '__main__':
    a = SCDebug()
    a.DEBUG = True
    a.iprint('INFO')




import unittest

from pysp.sjson import SJson


class InnerL2:
    def __init__(self, *args, **kwargs):
        self.msgs = ['hahaha', 'hohoho']
        self.dict = {
            'hello': 'hi'
        }


class InnerL1:
    def __init__(self, *args, **kwargs):
        self.msg = 'This is message.'
        self.nums = [1, 2]
        self.inner = InnerL2()


class OuterCls:
    def __init__(self, *args, **kwargs):
        self.a = 1.1
        self.b = 2.2
        self.c = 'text'
        self.inner = InnerL1()


class JsonSerialTest(unittest.TestCase):

    def test_json_serial(self):
        ocls = OuterCls()
        jstr = SJson.to_serial(ocls, indent=2)
        ojson = SJson.to_deserial(jstr, globals=globals())

        self.assertTrue(
            ocls.inner.inner.dict['hello'] == ojson.inner.inner.dict['hello'])
        self.assertTrue(
            str(ocls.inner.inner.msgs) == str(ojson.inner.inner.msgs))
        # print(globals())




import json
# import sys


class SJson:
    CLASS_MARK = '__cls__'

    @classmethod
    def to_serial(cls, o, **kwargs):
        indent = kwargs.get('indent', None)

        def default(o):
            if type(o) in [dict, list]:
                return o
            if hasattr(o, '__class__'):
                # return { **o.__dict__, '__class__': o.__class__.__name__}
                o.__dict__.update({cls.CLASS_MARK: o.__class__.__name__})
            return o.__dict__

        return json.dumps(default(o), indent=indent, default=default)

    @classmethod
    def to_deserial(cls, jstr, **kwargs):
        _globals = kwargs.get('globals', globals())

        def hook(o):
            if cls.CLASS_MARK in o:
                _cls = _globals[o[cls.CLASS_MARK]]
                # cls = getattr(sys.modules[__name__], o[cls.CLASS_MARK])
                del o[cls.CLASS_MARK]
                oc = _cls()
                oc.__dict__.update(o)
                return oc
            return o

        return json.loads(jstr, object_hook=hook)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("VERSION", 'r') as fd:
    version = fd.read().strip()

setup(
    name='pysp',
    version=version,
    description='Python-Support-Package',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='HyunSuk Lee',
    author_email='peanutstars.job@gmail.com',
    url='https://github.com/peanutstars/py-support-package',
    scripts=[],
    keywords='yaml configuration',
    packages=['pysp'],
    data_files = [('.', ['VERSION'])],
    include_package_data=True,
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Operating System Kernels :: Linux",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=[
        'PyYAML>=3.12',
        'SQLAlchemy>=1.2.14',
    ],
    setup_requires=[
    ],
)
